@@ -4,8 +4,8 @@ ASFPy methods
 Say some things about it here.
 """
 
-from pathlib import Path
 from operator import itemgetter
+import random
 import csv
 
 #################################################
@@ -13,39 +13,108 @@ import csv
 #
 # NOTE: These constants refer to data fields that
 #	are collected in forms, so may be changed
-#	accordingly.
+#	accordingly. The forms should also be
+#	preprocesseed to have the same labels.
+#
 #
 #################################################
 
+FLEXIBLE = "flexible"
 URM = "urm"
 LIM = "lim"
 SCHOOL = "du"
+NO_CONFLICTS_STATEMENT = "I AM NOT APPLYING TO ANY OF THE FOLLOWING PROFESSORS"
 
 #################################################
-# APPLICANT-ONLY METHODS
+# FILE HANDLERS
 #################################################
 
+def _id(pre, pad, i, sep = "_"):
+    """
+    Assign id of string concatenation: `pre + sep + pad + i`, for `i` a 0-padded
+    number, i.e., "002".
+    """
+    return pre + sep + str(i).rjust(pad, "0")
+
+def conflicts(faculty, no_conflict):
+    """
+    Extract faculty conflict of interest names and create set containing those
+    names. First remove the no_conflict statement, if any. Then return faculty.
+    """
+    faculty = set(faculty.split(", "))
+    return {fac for fac in faculty if no_conflict not in fac}
+
+def categories(categories):
+    """
+    Extract categories from stated categories of respondent.
+    """
+    return set(categories.split(", "))
+
+def get_element_by_id(_id, group_list):
+    """
+    Use next generator to find element in a group's list by identifier.
+    """
+    return next(e for e in group_list if _id == e["id"])
+
+def read_list_csv(filename):
+    """
+    Use Python's native CSV reader to load list.
+    """
+    with open(filename) as f:
+        loaded = [{k: v for k, v in row.items()}
+                for row in csv.DictReader(f, skipinitialspace=True)]
+    return loaded
+
 def read_preprocessed_editors_list_csv(filename):
     """
     Use Python's native CSV reader to load the editors list. Also,
     convert category stringlists to sets and endow an identifier.
     """
-    with open(filename) as f:
-        editors = [{k: v for k, v in row.items()}
-            for row in csv.DictReader(f, skipinitialspace=True)]
+    editors = read_list_csv(filename)
 
     n = 1
     for editor in editors:
-        editor["id"] = "EDI" + str(n).rjust(3, "0")
-        editor["categories"] = set(editor["categories"].split(", "))
+        editor["id"] = _id("EDI", 3, n)
+        editor["categories"] = categories(editor["categories"])
         editor["capacity"] = int(editor["capacity"])
         
         n += 1
 
     return editors
 
+def read_preprocessed_applicants_list_csv(filename):
+    """
+    Use Python's native CSV reader to load the applicant submissions list. Also,
+    convert category stringlists to sets and endow an identifier.
+    """
+    applicants = read_list_csv(filename)
+
+    n = 1
+    for applicant in applicants:
+        applicant["id"] = _id("APP", 3, n)
+        applicant["categories"] = categories(applicant["categories"])
+        applicant["conflicts"] = conflicts(applicant["conflicts"], NO_CONFLICTS_STATEMENT)
+        applicant[FLEXIBLE] = bool(applicant[FLEXIBLE])
+        applicant[URM] = bool(applicant[URM])
+        applicant[LIM] = bool(applicant[LIM])
+        applicant[SCHOOL] = bool(applicant[SCHOOL])
+
+        n += 1
+
+    return applicants
+
+def write_matchings_list_csv(matchings, filename):
+    """
+    Save dyad-format matchings to CSV list of matchings.
+    """
+    headers = matchings[0].keys()
+    with open(filename, 'w', newline='') as f:
+        w = csv.DictWriter(f, headers)
+        w.writeheader()
+        w.writerows(matchings)
+
 #################################################
-# APPLICANT-ONLY METHODS
+# APPLICANT PRIORITY METHODS
 #################################################
 
 def asfp_rank(applicant):
@@ -90,9 +159,17 @@ def asfp_rank(applicant):
 
     return rank
 
+def randomize(applicants):
+    """
+    Random permutation of applicant list. Typical usage is before
+    prioritization.
+    """
+    return random.sample(applicants, k = len(applicants))
+
 def prioritize(applicants, rank_method = asfp_rank):
     """
-    Prioritize applicants by rank of attributes.
+    Prioritize applicants by rank of attributes. Applicants are randomized
+    prior to running the ranking and sorting.
 
     Parameters
     ----------
@@ -110,6 +187,7 @@ def prioritize(applicants, rank_method = asfp_rank):
     """
     for a in applicants:
         a["rank"] = rank_method(a)
+
     return sorted(applicants, key = itemgetter("rank"))
 
 #################################################
@@ -198,6 +276,22 @@ def update_capacity(editor_id, editors):
         if editor["id"] == editor_id:
             editor["capacity"] -= 1
 
+def remove_conflicts(applicant, potential_editors):
+    """
+    Remove editors from potential editors who might be sources of conflict of
+    interest. These are typically by name, and not be assigned ASFPy _id
+    generator.
+    """
+    return [editor for editor in potential_editors if
+                editor["name"] not in applicant["conflicts"]]
+
+def find_potential_editors(applicant, editors):
+    """
+    Find potential editors given applicant categories and conflicts.
+    """
+    category_editors = editors_by_categories(editors, applicant["categories"])
+    return remove_conflicts(applicant, category_editors)
+
 def allocate(applicants, editors):
     """
     Allocate applicants to editors.
@@ -207,7 +301,7 @@ def allocate(applicants, editors):
 
     for applicant in applicants:
 
-        potential_editors = editors_by_categories(editors, applicant["categories"])
+        potential_editors = find_potential_editors(applicant, editors)
 
         if capacity(potential_editors) < 2:
         # If the editing capacity for an applicant is less than 2, continue to next applicant
@@ -234,7 +328,7 @@ def allocate(applicants, editors):
                 # If fewer than 2 student editors are available, skip to next
                 # applicant.
                     continue
-                elif applicant["is_flexible"]:
+                elif applicant[FLEXIBLE]:
                 # If an applicant is flexible and prefers to be matched with two
                 # student editors, find first match here.
                     student_editor_match = find_match(applicant, student_editors)
@@ -259,5 +353,50 @@ def allocate(applicants, editors):
 
     return {
         "matchings": matchings,
-        "unmatched": unmatched
+        "unmatched": unmatched,
+        "editors": editors
     }
+
+
+def format_matchings(matchings, applicants, editors):
+    """
+    Create dyadic format matchings, returning list of dicts with
+    pairings of editors and applicants, including categories and
+    statements with notes from applicants.
+
+    NOTE: The applicants and editors list args expect the original lists with
+        detailed information, not just the output of allocate.
+    """
+    dyads = []
+    for matching in matchings:
+        # Get applicant id from match and then get applicant details
+        # from applicants list
+        applicant_id = matching["applicant"]
+        applicant = get_element_by_id(applicant_id, applicants)
+
+        # Get editor ids and then get both detailed editor elements from list
+        editor_ids = matching["editors"]
+        for editor_id in editor_ids:
+            editor = get_element_by_id(editor_id, editors)
+
+            dyads.append({
+                "editor_id": editor["id"],
+                "editor_email": editor["email"],
+                "editor_categories": ", ".join(str(c) for c in editor["categories"]),
+                "applicant_id": applicant["id"],
+                "applicant_email": applicant["email"],
+                "applicant_statement": applicant["statement"],
+                "applicant_notes": applicant["notes"],
+                "applicant_categories": ", ".join(str(c) for c in applicant["categories"]),
+            })
+            
+    return dyads
+    
+def compile_unmatched_applicants(unmatched, applicants):
+    """
+    Compile list of unmatched applicants after allocation.
+
+    NOTE: unmatched is a list of applicant IDs and applicants in the official
+        list of applicants
+    """
+    return [a for a in applicants if a["id"] in unmatched]
